/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __STATISTICS_WIIU_H__
#define __STATISTICS_WIIU_H__
#include "config.h"

namespace WTF {

#define ENABLE_STATS_WIIU
#ifdef ENABLE_STATS_WIIU

/*
* Convenience functions to facilitate the setting and clearing of stats.  For more information and what
* the macro does, see the corresponding function in the StatisticsWiiU class. 
*/
#define STATS_WIIU_INC_PAINT_COUNT() StatisticsWiiU::instance()->incrementPaintCount()
#define STATS_WIIU_INC_IMAGE_COUNT() StatisticsWiiU::instance()->incrementImageCount()
#define STATS_WIIU_INC_FONTCACHE_COUNT() StatisticsWiiU::instance()->incrementFontCacheCount()
#define STATS_WIIU_INC_TRANSLAYER_COUNT() StatisticsWiiU::instance()->incrementTransparencyLayerCount()
#define STATS_WIIU_INC_IMAGEBUFFER_COUNT() StatisticsWiiU::instance()->incrementImageBufferCount()

#define STATS_WIIU_DEC_PAINT_COUNT() StatisticsWiiU::instance()->decrementPaintCount()
#define STATS_WIIU_DEC_IMAGE_COUNT() StatisticsWiiU::instance()->decrementImageCount()
#define STATS_WIIU_DEC_FONTCACHE_COUNT() StatisticsWiiU::instance()->decrementFontCacheCount()
#define STATS_WIIU_DEC_TRANSLAYER_COUNT() StatisticsWiiU::instance()->decrementTransparencyLayerCount()
#define STATS_WIIU_DEC_IMAGEBUFFER_COUNT() StatisticsWiiU::instance()->decrementImageBufferCount() 

#define STATS_WIIU_CLEAR_PAINT_COUNT() StatisticsWiiU::instance()->clearPaintCount() 
#define STATS_WIIU_CLEAR_IMAGE_COUNT() StatisticsWiiU::instance()->clearImageCount()
#define STATS_WIIU_CLEAR_FONTCACHE_COUNT() StatisticsWiiU::instance()->clearFontCacheCount()
#define STATS_WIIU_CLEAR_TRANSLAYER_COUNT() StatisticsWiiU::instance()->clearTransparencyLayerCount() 
#define STATS_WIIU_CLEAR_IMAGEBUFFER_COUNT() StatisticsWiiU::instance()->clearImageBufferCount() 
#define STATS_WIIU_CLEAR_ALL() StatisticsWiiU::instance()->clearAll()
 
#else

/*
* When the ENABLE_STATS_WIIU define is disabled, these macros become noops. 
*/
#define STATS_WIIU_INC_PAINT_COUNT()
#define STATS_WIIU_INC_IMAGE_COUNT()
#define STATS_WIIU_INC_FONTCACHE_COUNT()
#define STATS_WIIU_INC_TRANSLAYER_COUNT()
#define STATS_WIIU_INC_TRANSLAYER_COUNT()

#define STATS_WIIU_DEC_PAINT_COUNT()
#define STATS_WIIU_DEC_IMAGE_COUNT()
#define STATS_WIIU_DEC_FONTCACHE_COUNT()
#define STATS_WIIU_DEC_TRANSLAYER_COUNT()
#define STATS_WIIU_DEC_IMAGEBUFFER_COUNT()

#define STATS_WIIU_CLEAR_PAINT_COUNT() 
#define STATS_WIIU_CLEAR_IMAGE_COUNT()
#define STATS_WIIU_CLEAR_FONTCACHE_COUNT()
#define STATS_WIIU_CLEAR_TRANSLAYER_COUNT() 
#define STATS_WIIU_CLEAR_IMAGEBUFFER_COUNT() 
#define STATS_WIIU_CLEAR_ALL()
 
#endif

/*
* This class encapsulates statistics about webkit during it's execution. 
*/
class StatisticsWiiU {
	public:
	
		/*
		* Increments the paint count. 
		*/
		void incrementPaintCount(); 
		
		/*
		* Increments the image count.  Called when an image has been decoded and a texture has been created for that image. 
		*/
		void incrementImageCount(); 
		
		/*
		* Increments the font cache count.  This is called when a font cache is created for a new font. 
		*/
		void incrementFontCacheCount(); 
		
		/*
		* Increments the transparency layer count.  This is called when a begin transparency layer call is performed. 
		*/
		void incrementTransparencyLayerCount(); 
		
		/*
		* Increments the image buffer count.  This is called when a new image buffer is created and graphics context is attached to that image buffer. 
		*/
		void incrementImageBufferCount(); 
		
		/*
		* Decrements the paint count.  Currently not called anywhere.  Added for future proofing of code. 
		*/
		void decrementPaintCount(); 
		
		/*
		* Decrements the image count.  This is called when a frame for an image is cleared. 
		*/
		void decrementImageCount(); 
		
		/*
		* Decrements the font cache count.  This is called when a font cache has been destroyed for a font. 
		*/
		void decrementFontCacheCount(); 
		
		/*
		* Decrements the transparency layer count.  This is called when the end transparency layer call is performed. 
		*/
		void decrementTransparencyLayerCount(); 
		
		/*
		* Decrements the image buffer count.  This is called when the image buffer destructor is called. 
		*/
		void decrementImageBufferCount();
		
		/*
		* Clears the paint count and resets it to zero.  This is called as part of the clearAll call currently. 
		*/
		void clearPaintCount(); 
		
		/*
		* Clears the image count and resets it to zero.  This is called as part of the clearAll call currently. 
		*/
		void clearImageCount(); 
		
		/*
		* Clears the font cache count and resets it to zero.  This is called as part of the clearAll call currently. 
		*/
		void clearFontCacheCount(); 
		
		/*
		* Clears the transparency layer count and resets it to zero.  This is called as part of the clearAll call currently. 
		*/
		void clearTransparencyLayerCount();
		
		/*
		* Clears the image buffer count and resets it to zero.  This is called as part of the clearAll call currently.
		*/
		void clearImageBufferCount(); 
		
		/*
		* Clears all the current statistics and resets them all to zero. 
		*/
		void clearAll();
		
		/*
		* Returns the current number of paints since the page loaded successfully. 
		*/
		unsigned getPaintCount(); 
		
		/*
		* Returns the current number of images in use. 
		*/
		unsigned getImageCount(); 
		
		/*
		*  Return the current number of font caches that are allocated. 
		*/
		unsigned getFontCacheCount(); 
		
		/*
		* Returns the current number of transparency layers in use. 
		*/
		unsigned getTransparencyLayerCount(); 
		
		/*
		* Returns the current number of image buffers in use. 
		*/
		unsigned getImageBufferCount(); 
		
		/*
		* Returns the global instance to the StatisticsWiiU class. 
		*/
		static StatisticsWiiU *instance();
		
	private:
		StatisticsWiiU(); 
		static StatisticsWiiU *m_instance;
		unsigned m_paintCount;
		unsigned m_imageCount;
		unsigned m_fontCacheCount;
		unsigned m_transLayerCount; 
		unsigned m_imageBufferCount;
};

inline void StatisticsWiiU::incrementPaintCount() {
	m_paintCount++;
} 

inline void StatisticsWiiU::incrementImageCount() {
	m_imageCount++;
}

inline void StatisticsWiiU::incrementFontCacheCount() {
	m_fontCacheCount++;
} 

inline void StatisticsWiiU::incrementTransparencyLayerCount() {
	m_transLayerCount++;
} 

inline void StatisticsWiiU::incrementImageBufferCount() {
	m_imageBufferCount++;
}

inline void StatisticsWiiU::decrementPaintCount() {
	m_paintCount--;
} 

inline void StatisticsWiiU::decrementImageCount() {
	m_imageCount--;
}

inline void StatisticsWiiU::decrementFontCacheCount() {
	m_fontCacheCount--;
} 

inline void StatisticsWiiU::decrementTransparencyLayerCount() {
	m_transLayerCount--; 
} 

inline void StatisticsWiiU::decrementImageBufferCount() {
	m_imageBufferCount--;
}

};

#endif
