/*
 * Copyright (C) 2010, Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(WEB_AUDIO)

#include "NotImplemented.h"
#include <wtf/Platform.h> // PLATFORM(...)

#include <stdio.h> // snprintf

#include "AudioBus.h"

#include "AudioFileReader.h"
#include <wtf/PassOwnPtr.h>

#include <wkaudio/wkaudAudio.h>
#include <wkaudio/wkaudAudioPlatform.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#if PLATFORM(WIIU)
#include <shim.h>
#include <PlatformOS.h>
#endif

namespace WebCore {

namespace PrivateAudioUtils {

PassOwnPtr<AudioBus> audioBusFromInMemoryResource(const void *pData, u32 resourceSize) {
	WkSupAudioUtilsPrivate::PerChannelFloatSample smp;
	WkSupAudioUtilsPrivate::PerChannelFloatSample::AudioFormatError	errCode;
	errCode = smp.ConvertFromInMemoryData(pData, resourceSize);

	if (errCode == WkSupAudioUtilsPrivate::PerChannelFloatSample::eErr_Ok)
	{
		// create an AudioBus, but don't allocate any memory for samples
		OwnPtr<AudioBus> pBus = adoptPtr(new AudioBus(smp.mNumChannels, smp.mNumSamplesPerChannel, false));
		pBus->setSampleRate(smp.mSampleRate);
		pBus->setDataFormat(smp.mDataFormat);
		pBus->setLoopStartSamples( smp.mLoopStart );
		pBus->setLoopEndSamples(   smp.mLoopEnd );

		// copy the channel data pointers
		for(u32 nCh = 0; nCh < smp.mNumChannels; ++nCh) {
			pBus->set_s16ChannelData( nCh, smp.s16ChannelData[nCh], smp.mNumSamplesPerChannel );
			smp.s16ChannelData[nCh] = 0;	// transfer ownership, to keep the data from being deleted
		}

		return pBus.release();
	}

	return nullptr;
}

PassOwnPtr<AudioBus> loadWAVFileResource(const char *pFilename) {
	u32 resourceSize = 0;

	// Prepend standard path to content on disc if we don't have an absolute path already
	#if PLATFORM(WIIU)
	char openPath[512];
	#elif OS(UNIX)
	char openPath[FILENAME_MAX];
	#else
	char openPath[MAX_PATH];
	#endif
	if (pFilename[0] != '/') {
		#if PLATFORM(WIIU)
		strcpy(openPath, shim_GetFSContentDir());
		#else

		// This is the current working path, what we really want is the dir of the exe
		// which will not be the same if this is run from a command prompt.
		if (!getcwd(openPath, sizeof(openPath)))
			return nullptr;

		strcat(openPath, "\\data\\disc\\content\\");
		#endif
	} else
		openPath[0] = '\0';

	if (strlen(pFilename) + strlen(openPath) > (sizeof(openPath) - 1))
		return nullptr;

	strcat(openPath, pFilename);

	// Open...
	int fd = open(openPath, O_RDONLY, 0);
	if (fd < 0)
		return nullptr;

	// Get length...
	struct stat st;
	if (fstat(fd, &st) < 0)
		return nullptr;
	
	resourceSize = (u32)st.st_size;

	// Allocate memory 
	void *pData = malloc(st.st_size);
	if (pData == NULL) {
		close(fd);
		return nullptr;
	}
	
	// Read!
	if (read(fd, pData, st.st_size) < 0) {
		free(pData);
		return nullptr;
	}

	close(fd);

	OwnPtr<AudioBus> audioBus = audioBusFromInMemoryResource(pData, resourceSize);
	free(pData);
	return audioBus.release();
}

} // namespace PrivateAudioUtils

PassOwnPtr<AudioBus> AudioBus::loadPlatformResource(const char* name, double sampleRate) {
	char tmp[1024];
	snprintf(tmp, sizeof(tmp), "assets/webaudio/resources/%s.wav", name);
	OwnPtr<AudioBus> audioBus = PrivateAudioUtils::loadWAVFileResource(tmp);

	if (!audioBus.get())
		return nullptr;

//	if (audioBus->sampleRate() == sampleRate)
		return audioBus.release();

//	return AudioBus::createBySampleRateConverting(audioBus.get(), false, sampleRate);
}


PassOwnPtr<AudioBus> createBusFromInMemoryAudioFile(const void* data, size_t dataSize, bool mixToMono, double sampleRate) {
	OwnPtr<AudioBus> audioBus;
	audioBus = PrivateAudioUtils::audioBusFromInMemoryResource(data, dataSize);

	if (audioBus.get())
	{
		if (mixToMono && (audioBus->numberOfChannels() > 1))
		{
			// mix stereo to mono
			s16* p0 = audioBus->get_s16ChannelData( 0 );
			s16* p1 = audioBus->get_s16ChannelData( 1 );
			unsigned len = audioBus->length();
			s32 mix;

			// average both channels into channel 0
			for (unsigned i=0; i<len; i++)
			{
				mix = (s32)*p0 + (s32)*p1++;
				*p0++ = (s16)(mix / 2);
			}

			// wipe the audio samples from the processor cache
			LibPlatform::OS::StoreRange( audioBus->get_s16ChannelData( 0 ), sizeof(s16) * len );

			// Set the number of channels to 1 but leave the data in place.
			// Freeing the data here would produce an error, but doing this
			// will result in just one copy and just one AX voice assigned.
			audioBus->setNumberOfChannels( 1 );
		}
	}
	else
	{
		// If everything has failed we have to create a dummy bus so the upstream won't fail.
		audioBus = adoptPtr(new AudioBus(mixToMono ? 1 : 2, 128, true));
		audioBus->zero();
		audioBus->setSampleRate(sampleRate);

		// This is the only way we know the decoding failed, so set the error value here.
		audioBus->setErrorValue( WkSupAudioUtilsPrivate::PerChannelFloatSample::eErr_BadFormat );
	}

    return audioBus.release();
}

} // namespace WebCore

#endif // ENABLE(WEB_AUDIO)
