/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if USE(ACCELERATED_COMPOSITING)

#include "LayerWiiU.h"

#include "TextStream.h"
#include "FloatPoint.h"
#include "FloatQuad.h"
#include "FloatRect.h"
#include "IntRect.h"
#include "TransformationMatrix.h"
#include "wtf/text/CString.h"
#include <wtf/Deque.h>

namespace WebCore {

using namespace std;

static int s_nextLayerId = 1;
const float layerEpsilon = 1e-4f;

void ResetNextLayerID()
{
    s_nextLayerId = 1;
}

struct HomogeneousCoordinate {
    HomogeneousCoordinate(double newX, double newY, double newZ, double newW)
        : x(newX)
        , y(newY)
        , z(newZ)
        , w(newW)
    {
    }

    bool shouldBeClipped() const
    {
        return w <= 0;
    }

    FloatPoint cartesianPoint2d() const
    {
        if (w == 1)
            return FloatPoint(x, y);

        // For now, because this code is used privately only by CCMathUtil, it should never be called when w == 0, and we do not yet need to handle that case.
        ASSERT(w);
        double invW = 1.0 / w;
        return FloatPoint(x * invW, y * invW);
    }

    double x;
    double y;
    double z;
    double w;
};

static HomogeneousCoordinate projectPoint(const TransformationMatrix& transform, const FloatPoint& p)
{
    double x = p.x();
    double y = p.y();
    double z = -(transform.m13() * x + transform.m23() * y + transform.m43()) / transform.m33();
    // implicit definition of w = 1;

    double outX = x * transform.m11() + y * transform.m21() + z * transform.m31() + transform.m41();
    double outY = x * transform.m12() + y * transform.m22() + z * transform.m32() + transform.m42();
    double outZ = x * transform.m13() + y * transform.m23() + z * transform.m33() + transform.m43();
    double outW = x * transform.m14() + y * transform.m24() + z * transform.m34() + transform.m44();

    return HomogeneousCoordinate(outX, outY, outZ, outW);
}

static HomogeneousCoordinate mapPoint(const TransformationMatrix& transform, const FloatPoint& p)
{
    double x = p.x();
    double y = p.y();
    // implicit definition of z = 0;
    // implicit definition of w = 1;

    double outX = x * transform.m11() + y * transform.m21() + transform.m41();
    double outY = x * transform.m12() + y * transform.m22() + transform.m42();
    double outZ = x * transform.m13() + y * transform.m23() + transform.m43();
    double outW = x * transform.m14() + y * transform.m24() + transform.m44();

    return HomogeneousCoordinate(outX, outY, outZ, outW);
}

static HomogeneousCoordinate computeClippedPointForEdge(const HomogeneousCoordinate& h1, const HomogeneousCoordinate& h2)
{
    // Points h1 and h2 form a line in 4d, and any point on that line can be represented
    // as an interpolation between h1 and h2:
    //    p = (1-t) h1 + (t) h2
    //
    // We want to compute point p such that p.w == epsilon, where epsilon is a small
    // non-zero number. (but the smaller the number is, the higher the risk of overflow)
    // To do this, we solve for t in the following equation:
    //    p.w = epsilon = (1-t) * h1.w + (t) * h2.w
    //
    // Once paramter t is known, the rest of p can be computed via p = (1-t) h1 + (t) h2.

    // Technically this is a special case of the following assertion, but its a good idea to keep it an explicit sanity check here.
    ASSERT(h2.w != h1.w);
    // Exactly one of h1 or h2 (but not both) must be on the negative side of the w plane when this is called.
    ASSERT(h1.shouldBeClipped() ^ h2.shouldBeClipped());

    double w = 0.00001; // or any positive non-zero small epsilon

    double t = (w - h1.w) / (h2.w - h1.w);

    double x = (1-t) * h1.x + t * h2.x;
    double y = (1-t) * h1.y + t * h2.y;
    double z = (1-t) * h1.z + t * h2.z;

    return HomogeneousCoordinate(x, y, z, w);
}

static inline void expandBoundsToIncludePoint(float& xmin, float& xmax, float& ymin, float& ymax, const FloatPoint& p)
{
    xmin = std::min(p.x(), xmin);
    xmax = std::max(p.x(), xmax);
    ymin = std::min(p.y(), ymin);
    ymax = std::max(p.y(), ymax);
}

static inline void addVertexToClippedQuad(const FloatPoint& newVertex, FloatPoint clippedQuad[8], int& numVerticesInClippedQuad)
{
    clippedQuad[numVerticesInClippedQuad] = newVertex;
    numVerticesInClippedQuad++;
}

void mapClippedQuad(const TransformationMatrix& transform, const FloatQuad& srcQuad, FloatPoint clippedQuad[8], int& numVerticesInClippedQuad)
{
    HomogeneousCoordinate h1 = mapPoint(transform, srcQuad.p1());
    HomogeneousCoordinate h2 = mapPoint(transform, srcQuad.p2());
    HomogeneousCoordinate h3 = mapPoint(transform, srcQuad.p3());
    HomogeneousCoordinate h4 = mapPoint(transform, srcQuad.p4());

    // The order of adding the vertices to the array is chosen so that clockwise / counter-clockwise orientation is retained.

    numVerticesInClippedQuad = 0;

    if (!h1.shouldBeClipped())
        addVertexToClippedQuad(h1.cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (h1.shouldBeClipped() ^ h2.shouldBeClipped())
        addVertexToClippedQuad(computeClippedPointForEdge(h1, h2).cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (!h2.shouldBeClipped())
        addVertexToClippedQuad(h2.cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (h2.shouldBeClipped() ^ h3.shouldBeClipped())
        addVertexToClippedQuad(computeClippedPointForEdge(h2, h3).cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (!h3.shouldBeClipped())
        addVertexToClippedQuad(h3.cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (h3.shouldBeClipped() ^ h4.shouldBeClipped())
        addVertexToClippedQuad(computeClippedPointForEdge(h3, h4).cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (!h4.shouldBeClipped())
        addVertexToClippedQuad(h4.cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    if (h4.shouldBeClipped() ^ h1.shouldBeClipped())
        addVertexToClippedQuad(computeClippedPointForEdge(h4, h1).cartesianPoint2d(), clippedQuad, numVerticesInClippedQuad);

    ASSERT(numVerticesInClippedQuad <= 8);
}

FloatRect computeEnclosingRectOfVertices(FloatPoint vertices[], int numVertices)
{
    if (numVertices < 2)
        return FloatRect();

    float xmin = std::numeric_limits<float>::max();
    float xmax = std::numeric_limits<float>::min();
    float ymin = std::numeric_limits<float>::max();
    float ymax = std::numeric_limits<float>::min();

    for (int i = 0; i < numVertices; ++i)
        expandBoundsToIncludePoint(xmin, xmax, ymin, ymax, vertices[i]);

    return FloatRect(FloatPoint(xmin, ymin), FloatSize(xmax - xmin, ymax - ymin));
}

inline static float perpProduct(const FloatSize& u, const FloatSize& v)
{
    return u.width() * v.height() - u.height() * v.width();
}

// Tests if two edges defined by their endpoints (a,b) and (c,d) intersect. Returns true and the
// point of intersection if they do and false otherwise.
static bool edgeEdgeTest(const FloatPoint& a, const FloatPoint& b, const FloatPoint& c, const FloatPoint& d, FloatPoint& r)
{
    FloatSize u = b - a;
    FloatSize v = d - c;
    FloatSize w = a - c;

    float denom = perpProduct(u, v);

    // If denom == 0 then the edges are parallel.
    if (!denom)
        return false;

    float s = perpProduct(v, w) / denom;
    if (s < 0 || s > 1)
        return false;

    float t = perpProduct(u, w) / denom;
    if (t < 0 || t > 1)
        return false;

    u.scale(s);
    r = a + u;
    return true;
}

static float const checkFloatingPointNumericAccuracy(float a, float b)
{
    float absDif = std::abs(b - a);
    float absMax = std::max(std::abs(b), std::abs(a));
    // Check to see if we've got a result with a reasonable amount of error.
    return absDif / absMax;
}

void LayerSorterWiiU::addLayer(RefPtr<LayerWiiU> layer)
{
    m_sortedList.append(layer);
}

void LayerSorterWiiU::sort()
{
    sort(m_sortedList.begin(), m_sortedList.end());
}

// Checks whether layer "a" draws on top of layer "b". The weight value returned is an indication of
// the maximum z-depth difference between the layers or zero if the layers are found to be intesecting
// (some features are in front and some are behind).
LayerSorterWiiU::ABCompareResult LayerSorterWiiU::checkOverlap(LayerShape* a, LayerShape* b, float zThreshold, float& weight)
{
    weight = 0;

    // Early out if the projected bounds don't overlap.
   if (!a->projectedBounds.intersects(b->projectedBounds))
        return None;

    FloatPoint aPoints[4] = {a->projectedQuad.p1(), a->projectedQuad.p2(), a->projectedQuad.p3(), a->projectedQuad.p4() };
    FloatPoint bPoints[4] = {b->projectedQuad.p1(), b->projectedQuad.p2(), b->projectedQuad.p3(), b->projectedQuad.p4() };

    // Make a list of points that inside both layer quad projections.
    Vector<FloatPoint> overlapPoints;

    // Check all four corners of one layer against the other layer's quad.
    for (int i = 0; i < 4; ++i) {
        if (a->projectedQuad.containsPoint(bPoints[i])) {
            overlapPoints.append(bPoints[i]);
        }
        if (b->projectedQuad.containsPoint(aPoints[i])) {
            overlapPoints.append(aPoints[i]);
        }
    }

    // Check all the edges of one layer for intersection with the other layer's edges.
    FloatPoint r;
    for (int ea = 0; ea < 4; ++ea)
        for (int eb = 0; eb < 4; ++eb)
            if (edgeEdgeTest(aPoints[ea], aPoints[(ea + 1) % 4],
                             bPoints[eb], bPoints[(eb + 1) % 4],
                             r)) {
                overlapPoints.append(r);
            }

    if (!overlapPoints.size())
        return None;

    // Check the corresponding layer depth value for all overlap points to determine
    // which layer is in front.
    float maxPositive = 0;
    float maxNegative = 0;
    bool accurate = false;

    for (unsigned o = 0; o < overlapPoints.size(); o++) {
        float za = a->layerZFromProjectedPoint(overlapPoints[o]);
        float zb = b->layerZFromProjectedPoint(overlapPoints[o]);

        if (checkFloatingPointNumericAccuracy(za, zb) > layerEpsilon)
          accurate = true;

        float diff = 0;

        diff = za - zb;

        if (diff > maxPositive)
            maxPositive = diff;
        if (diff < maxNegative)
            maxNegative = diff;
    }

    if (!accurate)
      return ABeforeB;

    float maxDiff = (fabsf(maxPositive) > fabsf(maxNegative) ? maxPositive : maxNegative);
   // OS::Report("maxDiff %f\n", maxDiff);
    // If the results are inconsistent (and the z difference substantial to rule out
    // numerical errors) then the layers are intersecting. We will still return an
    // order based on the maximum depth difference but with an edge weight of zero
    // these layers will get priority if a graph cycle is present and needs to be broken.
    if (maxPositive > zThreshold && maxNegative < -zThreshold)
        weight = 0;
    else
        weight = fabsf(maxDiff);

    // Maintain relative order if the layers have the same depth at all intersection points.
   if (maxDiff <= 0)
        return ABeforeB;

    return BBeforeA;
}

LayerSorterWiiU::LayerShape::LayerShape(float width, float height, const TransformationMatrix& drawTransform)
{

    FloatQuad layerQuad(FloatRect(0, 0, width, height));
    // Compute the projection of the layer quad onto the z = 0 plane.

    FloatPoint clippedQuad[8];
    int numVerticesInClippedQuad;
    mapClippedQuad(drawTransform, layerQuad, clippedQuad, numVerticesInClippedQuad);

    if (numVerticesInClippedQuad < 3) {
        projectedBounds = FloatRect();
        return;
    }

    projectedBounds = computeEnclosingRectOfVertices(clippedQuad, numVerticesInClippedQuad);

    // NOTE: it will require very significant refactoring and overhead to deal with
    // generalized polygons or multiple quads per layer here. For the sake of layer
    // sorting it is equally correct to take a subsection of the polygon that can be made
    // into a quad. This will only be incorrect in the case of intersecting layers, which
    // are not supported yet anyway.
    projectedQuad.setP1(clippedQuad[0]);
    projectedQuad.setP2(clippedQuad[1]);
    projectedQuad.setP3(clippedQuad[2]);
    if (numVerticesInClippedQuad >= 4)
        projectedQuad.setP4(clippedQuad[3]);
    else
        projectedQuad.setP4(clippedQuad[2]); // this will be a degenerate quad that is actually a triangle.

    // Compute the normal of the layer's plane.
    FloatPoint3D c1 = drawTransform.mapPoint(FloatPoint3D(0, 0, 0));
    FloatPoint3D c2 = drawTransform.mapPoint(FloatPoint3D(0, 1, 0));
    FloatPoint3D c3 = drawTransform.mapPoint(FloatPoint3D(1, 0, 0));
    FloatPoint3D c12 = c2 - c1;
    FloatPoint3D c13 = c3 - c1;
    layerNormal = c13.cross(c12);

    transformOrigin = c1;
}


float LayerSorterWiiU::LayerShape::layerZFromProjectedPoint(const FloatPoint& p) const
{
    const FloatPoint3D zAxis(0, 0, 1);
    FloatPoint3D w = FloatPoint3D(p) - transformOrigin;

    float d = layerNormal.dot(zAxis);
    float n = -layerNormal.dot(w);

    // Check if layer is parallel to the z = 0 axis which will make it
    // invisible and hence returning zero is fine.
    if (!d)
        return 0;
  //  OS::Report("n/d: %f/%f = %f", n, d, n/d);
    // The intersection point would be given by:
    // p + (n / d) * u  but since we are only interested in the
    // z coordinate and p's z coord is zero, all we need is the value of n/d.
    return n / d;
}

void LayerSorterWiiU::createGraphNodes(LayerList::iterator first, LayerList::iterator last)
{

    float minZ = FLT_MAX;
    float maxZ = -FLT_MAX;
    for (LayerList::const_iterator it = first; it < last; it++) {
        m_nodes.append(GraphNode(it->get()));
        GraphNode& node = m_nodes.at(m_nodes.size() - 1);
        if (!node.layer->drawsContent())
            continue;

        TransformationMatrix drawTransform;
        float layerWidth, layerHeight;


        drawTransform = node.layer->drawTransform();

       float x = node.layer->position().x();
       float y = node.layer->position().y();

       layerWidth = node.layer->bounds().width();
       layerHeight = node.layer->bounds().height();

        node.shape = LayerShape(layerWidth, layerHeight, drawTransform);

        maxZ = max(maxZ, node.shape.transformOrigin.z());
        minZ = min(minZ, node.shape.transformOrigin.z());
    }
}

void LayerSorterWiiU::createGraphEdges()
{

    // Fraction of the total zRange below which z differences
    // are not considered reliable.
    const float zThresholdFactor = 0.01;
    float zThreshold = m_zRange * zThresholdFactor;

    for (unsigned na = 0; na < m_nodes.size(); na++) {
        GraphNode& nodeA = m_nodes[na];
        if (!nodeA.layer->drawsContent())
            continue;
        for (unsigned nb = na + 1; nb < m_nodes.size(); nb++) {
            GraphNode& nodeB = m_nodes[nb];
            if (!nodeB.layer->drawsContent())
                continue;
            float weight = 0;
            ABCompareResult overlapResult = checkOverlap(&nodeA.shape, &nodeB.shape, zThreshold, weight);

            GraphNode* startNode = 0;
            GraphNode* endNode = 0;
            if (overlapResult == ABeforeB) {
                startNode = &nodeA;
                endNode = &nodeB;
            } else if (overlapResult == BBeforeA) {
                startNode = &nodeB;
                endNode = &nodeA;
            }

            if (startNode) {
              m_edges.append(GraphEdge(startNode, endNode, weight));
            }
        }
    }

    for (unsigned i = 0; i < m_edges.size(); i++) {
        GraphEdge& edge = m_edges[i];
        m_activeEdges.add(&edge, &edge);
        edge.from->outgoing.append(&edge);
        edge.to->incoming.append(&edge);
        edge.to->incomingEdgeWeight += edge.weight;
    }
}

// Finds and removes an edge from the list by doing a swap with the
// last element of the list.
void LayerSorterWiiU::removeEdgeFromList(GraphEdge* edge, Vector<GraphEdge*>& list)
{
    size_t edgeIndex = list.find(edge);
    ASSERT(edgeIndex != notFound);
    if (list.size() == 1) {
        ASSERT(!edgeIndex);
        list.clear();
        return;
    }
    if (edgeIndex != list.size() - 1)
        list[edgeIndex] = list[list.size() - 1];

    list.removeLast();
}

// Sorts the given list of layers such that they can be painted in a back-to-front
// order. Sorting produces correct results for non-intersecting layers that don't have
// cyclical order dependencies. Cycles and intersections are broken aribtrarily.
// Sorting of layers is done via a topological sort of a directed graph whose nodes are
// the layers themselves. An edge from node A to node B signifies that layer A needs to
// be drawn before layer B. If A and B have no dependency between each other, then we
// preserve the ordering of those layers as they were in the original list.
//
// The draw order between two layers is determined by projecting the two triangles making
// up each layer quad to the Z = 0 plane, finding points of intersection between the triangles
// and backprojecting those points to the plane of the layer to determine the corresponding Z
// coordinate. The layer with the lower Z coordinate (farther from the eye) needs to be rendered
// first.
//
// If the layer projections don't intersect, then no edges (dependencies) are created
// between them in the graph. HOWEVER, in this case we still need to preserve the ordering
// of the original list of layers, since that list should already have proper z-index
// ordering of layers.
//
void LayerSorterWiiU::sort(LayerList::iterator first, LayerList::iterator last)
{
    createGraphNodes(first, last);

    createGraphEdges();

    Vector<GraphNode*> sortedList;
    Deque<GraphNode*> noIncomingEdgeNodeList;

    // Find all the nodes that don't have incoming edges.
    for (NodeList::iterator la = m_nodes.begin(); la < m_nodes.end(); la++) {
        if (!la->incoming.size())
            noIncomingEdgeNodeList.append(la);
    }
    while (m_activeEdges.size() || noIncomingEdgeNodeList.size()) {
        while (noIncomingEdgeNodeList.size()) {

            // It is necessary to preserve the existing ordering of layers, when there are
            // no explicit dependencies (because this existing ordering has correct
            // z-index/layout ordering). To preserve this ordering, we process Nodes in
            // the same order that they were added to the list.
            GraphNode* fromNode = noIncomingEdgeNodeList.takeFirst();
            // Add it to the final list.
            sortedList.append(fromNode);

            // Remove all its outgoing edges from the graph.
            for (unsigned i = 0; i < fromNode->outgoing.size(); i++) {
                GraphEdge* outgoingEdge = fromNode->outgoing[i];

                m_activeEdges.remove(outgoingEdge);
                removeEdgeFromList(outgoingEdge, outgoingEdge->to->incoming);
                outgoingEdge->to->incomingEdgeWeight -= outgoingEdge->weight;

                if (!outgoingEdge->to->incoming.size())
                    noIncomingEdgeNodeList.append(outgoingEdge->to);
            }
            fromNode->outgoing.clear();
        }

        if (!m_activeEdges.size())
            break;

        // If there are still active edges but the list of nodes without incoming edges
        // is empty then we have run into a cycle. Break the cycle by finding the node
        // with the smallest overall incoming edge weight and use it. This will favor
        // nodes that have zero-weight incoming edges i.e. layers that are being
        // occluded by a layer that intersects them.
        float minIncomingEdgeWeight = FLT_MAX;
        GraphNode* nextNode = 0;
        for (unsigned i = 0; i < m_nodes.size(); i++) {
            if (m_nodes[i].incoming.size() && m_nodes[i].incomingEdgeWeight < minIncomingEdgeWeight) {
                minIncomingEdgeWeight = m_nodes[i].incomingEdgeWeight;
                nextNode = &m_nodes[i];
            }
        }
        ASSERT(nextNode);
        // Remove all its incoming edges.
        for (unsigned e = 0; e < nextNode->incoming.size(); e++) {
            GraphEdge* incomingEdge = nextNode->incoming[e];

            m_activeEdges.remove(incomingEdge);
            removeEdgeFromList(incomingEdge, incomingEdge->from->outgoing);
        }
        nextNode->incoming.clear();
        nextNode->incomingEdgeWeight = 0;
        noIncomingEdgeNodeList.append(nextNode);

    }

    // Note: The original elements of the list are in no danger of having their ref count go to zero
    // here as they are all nodes of the layer hierarchy and are kept alive by their parent nodes.
    int count = 0;
    for (LayerList::iterator it = first; it < last; it++)
        *it = sortedList[count++]->layer;


    m_nodes.clear();
    m_edges.clear();
    m_activeEdges.clear();
}


PassRefPtr<LayerWiiU> LayerWiiU::create(GraphicsLayerWiiU* owner)
{
    return adoptRef(new LayerWiiU(owner));
}

LayerWiiU::LayerWiiU(GraphicsLayerWiiU* owner)
    : m_owner(owner)
    , m_contentsDirty(false)
    , m_layerId(s_nextLayerId++)
    , m_parent(0)
    , m_anchorPoint(0.5, 0.5)
    , m_backgroundColor(0, 0, 0, 0)
    , m_debugBorderWidth(0)
    , m_opacity(1.0)
    , m_zPosition(0.0)
    , m_anchorPointZ(0)
    , m_masksToBounds(false)
    , m_opaque(true)
    , m_needsDisplayOnBoundsChange(false)
    , m_doubleSided(true)
    , m_isRootLayer(false)
    , m_isTransformLayer(false)
    , m_replicaLayer(0)
    , m_preserves3D(false)
    , m_minX(0.0)
    , m_maxX(0.0)
    , m_minY(0.0)
    , m_maxY(0.0)
    , m_minZ(0.0)
    , m_maxZ(0.0)
    , m_drawOpacity(0)
{
}

LayerWiiU::~LayerWiiU()
{
    // Our parent should be holding a reference to us so there should be no
    // way for us to be destroyed while we still have a parent.
    ASSERT(!parent());

    // Remove the parent reference from all children.
    removeAllChildren();
}

void LayerWiiU::cleanupResources()
{
}

void LayerWiiU::cleanupResourcesRecursive()
{
    for (size_t i = 0; i < children().size(); ++i)
        children()[i]->cleanupResourcesRecursive();

    if (maskLayer())
        maskLayer()->cleanupResourcesRecursive();
    if (replicaLayer())
        replicaLayer()->cleanupResourcesRecursive();

    cleanupResources();
}

void LayerWiiU::setNeedsCommit()
{
    // Call notifySyncRequired(), which for non-root layers plumbs through to
    // call setRootLayerNeedsDisplay() on the WebView, which will cause LayerRendererWiiU
    // to render a frame.
    // This function has no effect on root layers.
    if (m_owner)
        m_owner->notifySyncRequired();
}

void LayerWiiU::setParent(LayerWiiU* layer)
{
    ASSERT(!layer || !layer->hasAncestor(this));
    m_parent = layer;
}

bool LayerWiiU::hasAncestor(LayerWiiU* ancestor) const
{
    for (LayerWiiU* layer = parent(); layer; layer = layer->parent()) {
        if (layer == ancestor)
            return true;
    }
    return false;
}

void LayerWiiU::addChild(PassRefPtr<LayerWiiU> child)
{
    insertChild(child, numChildren());
}

void LayerWiiU::insertChild(PassRefPtr<LayerWiiU> child, size_t index)
{
    index = min(index, m_children.size());
    child->removeFromParent();
    child->setParent(this);
    m_children.insert(index, child);
    setNeedsCommit();
}

void LayerWiiU::removeFromParent()
{
    if (m_parent)
        m_parent->removeChild(this);
}

void LayerWiiU::removeChild(LayerWiiU* child)
{
    int foundIndex = indexOfChild(child);
    if (foundIndex == -1)
        return;

    child->setParent(0);
    m_children.remove(foundIndex);
    setNeedsCommit();
}

void LayerWiiU::replaceChild(LayerWiiU* reference, PassRefPtr<LayerWiiU> newLayer)
{
    ASSERT_ARG(reference, reference);
    ASSERT_ARG(reference, reference->parent() == this);

    if (reference == newLayer)
        return;

    int referenceIndex = indexOfChild(reference);
    if (referenceIndex == -1) {
        ASSERT_NOT_REACHED();
        return;
    }

    reference->removeFromParent();

    if (newLayer) {
        newLayer->removeFromParent();
        insertChild(newLayer, referenceIndex);
    }
}

int LayerWiiU::indexOfChild(const LayerWiiU* reference)
{
    for (size_t i = 0; i < m_children.size(); i++) {
        if (m_children[i] == reference)
            return i;
    }
    return -1;
}

void LayerWiiU::setBounds(const IntSize& size)
{
    if (bounds() == size)
        return;

    bool firstResize = !bounds().width() && !bounds().height() && size.width() && size.height();

    m_bounds = size;

    if (firstResize)
        setNeedsDisplay(FloatRect(0, 0, bounds().width(), bounds().height()));
    else
        setNeedsCommit();
}


const LayerWiiU* LayerWiiU::rootLayer() const
{
    const LayerWiiU* layer = this;
    for (LayerWiiU* parent = layer->parent(); parent; layer = parent, parent = parent->parent()) { }
    return layer;
}

void LayerWiiU::removeAllChildren()
{
    while (m_children.size()) {
        LayerWiiU* layer = m_children[0].get();
        ASSERT(layer->parent());
        layer->removeFromParent();
    }
}

void LayerWiiU::setChildren(const Vector<RefPtr<LayerWiiU> >& children)
{
    if (children == m_children)
        return;

    removeAllChildren();
    size_t listSize = children.size();
    for (size_t i = 0; i < listSize; i++)
        addChild(children[i]);
}

LayerWiiU* LayerWiiU::parent() const
{
    return m_parent;
}

void LayerWiiU::setName(const String& name)
{
    m_name = name;
}

void LayerWiiU::setNeedsDisplay(const FloatRect& dirtyRect)
{
    // Simply mark the contents as dirty. For non-root layers, the call to
    // setNeedsCommit will schedule a fresh compositing pass.
    // For the root layer, setNeedsCommit has no effect.
    m_contentsDirty = true;

    m_dirtyRect.unite(dirtyRect);
    setNeedsCommit();
}

void LayerWiiU::setNeedsDisplay()
{
    m_dirtyRect.setLocation(FloatPoint());
    m_dirtyRect.setSize(bounds());
    m_contentsDirty = true;
    setNeedsCommit();
}

void LayerWiiU::resetNeedsDisplay()
{
    m_dirtyRect = FloatRect();
    m_contentsDirty = false;
}

String LayerWiiU::layerTreeAsText() const
{
    TextStream ts;
    dumpLayer(ts, 0);
    return ts.release();
}

static void writeIndent(TextStream& ts, int indent)
{
    for (int i = 0; i != indent; ++i)
        ts << "  ";
}

void LayerWiiU::dumpLayer(TextStream& ts, int indent) const
{
    writeIndent(ts, indent);
    ts << layerTypeAsString() << "(" << m_name << ")\n";
    dumpLayerProperties(ts, indent+2);
    if (m_replicaLayer) {
        writeIndent(ts, indent+2);
        ts << "Replica:\n";
        m_replicaLayer->dumpLayer(ts, indent+3);
    }
    if (m_maskLayer) {
        writeIndent(ts, indent+2);
        ts << "Mask:\n";
        m_maskLayer->dumpLayer(ts, indent+3);
    }
    for (size_t i = 0; i < m_children.size(); ++i)
        m_children[i]->dumpLayer(ts, indent+1);
}

void LayerWiiU::dumpLayerProperties(TextStream& ts, int indent) const
{
    writeIndent(ts, indent);
    ts << "id: " << id() << " drawsContent: " << drawsContent() << " bounds " << m_bounds.width() << "x" << m_bounds.height() << "\n";

}

void LayerWiiU::setBorderColor(const Color& color)
{
    m_debugBorderColor = color;
    setNeedsCommit();
}

void LayerWiiU::setBorderWidth(float width)
{
    m_debugBorderWidth = width;
    setNeedsCommit();
}

void LayerWiiU::drawDebugBorder(GraphicsContext *ctx, float width, float height)
{
    setBorderColor(Color(255, 0, 255));
    setBorderWidth(5);

    ctx->save();
    ctx->setStrokeColor(m_debugBorderColor, ColorSpaceDeviceRGB);
    ctx->strokeRect(FloatRect(0, 0, width, height), m_debugBorderWidth);

    if (m_owner->showRepaintCounter()) {

    } 

    ctx->restore();
}

void LayerWiiU::paintContentsIfDirty()
{
	paintContentLayerIfDirty(); 	    
	
	//recursively paint any children
    for (int i = 0; i < children().size(); ++i) {
        children()[i]->paintContentsIfDirty();
    }
}


void LayerWiiU::paint(GraphicsContext *ctx, bool preserves3D, float opacity)
{
    ASSERT(ctx);

  //  float width  = (float)contentBounds().width();
 //   float height = (float)contentBounds().height();

  //  OS::Report("Layer %s Pos (%fx%f)  Size (%fx%f)\n", name().utf8().data(), m_owner->position().x(), m_owner->position().y(), width, height);

    if (m_owner->preserves3D())
        preserves3D = true;

    if (preserves3D) {
        LayerSorterWiiU  sorter;
        //Process transforms based on stacking order.
        process3DTransforms(ctx, &sorter, opacity);

        sorter.sort();
        for (LayerSorterWiiU::LayerList::iterator i = sorter.m_sortedList.begin(); i < sorter.m_sortedList.end(); i++)
        {
            (*i)->paint3D(ctx);
        }
    } else {
        paint2D(ctx, opacity);
    }
}

void LayerWiiU::paint2D(GraphicsContext *ctx, float opacity)
{

    float anchor_x = m_owner->anchorPoint().x();
    float anchor_y = m_owner->anchorPoint().y();
    float width  = (float)contentBounds().width();
    float height = (float)contentBounds().height();
    float pos_x = m_owner->position().x();
    float pos_y = m_owner->position().y();
    ctx->save();

    if (masksToBounds()) {
        ctx->clip(FloatRect(pos_x, pos_y, width, height));
    }


  //  OS::Report("Paint 2D Layer %s Pos (%fx%f)  Size (%fx%f)\n", name().utf8().data(), m_owner->position().x(), m_owner->position().y(), position().x(), position().y());
    TransformationMatrix t = transform();

    //Flatten Transform
    t.setM31(0);
    t.setM32(0);
    

    ctx->translate(anchor_x*width + pos_x, anchor_y*height + pos_y);

    //concatenate our transform so that is is applied to any children
    ctx->concatCTM(t);

    // Apply perspective Transform
    if (!sublayerTransform().isIdentity()) {
        ctx->concatCTM(sublayerTransform());
    }
    //translate back from anchor
    ctx->translate(-anchor_x*width, -anchor_y*height);

    m_accumulatedOpacity = opacity * m_opacity;

    if (drawsContent() && (width > 0) && (height > 0))  {

        // If backface visablity is hidden and the m33 is negative the back then 
		// We set the alpha to 0 to make it invisable. 
        if (!m_doubleSided && ctx->getCTMAsTransformationMatrix().inverse().m33() < 0)
            ctx->setAlpha(0);
        else
            ctx->setAlpha(m_accumulatedOpacity);
        paintContents(ctx);
     }

    //recursively paint any children
    for (int i = 0; i < children().size(); ++i) {
        children()[i]->paint(ctx, false, m_accumulatedOpacity);
    }
    ctx->restore();

}

void LayerWiiU::process3DTransforms(GraphicsContext *ctx, LayerSorterWiiU * sorter, float opacity)
{
    float anchor_x = m_owner->anchorPoint().x();
    float anchor_y = m_owner->anchorPoint().y();

    float x = m_owner->position().x();
    float y = m_owner->position().y();
    float width  = (float)contentBounds().width();
    float height = (float)contentBounds().height();

  //  int id = m_layerId;
 //  OS::Report("Transforming Layer(%d) %s Pos (%fx%f)  Size (%fx%f)\n", id, name().utf8().data(), m_owner->position().x(), m_owner->position().y(), width, height);

    ctx->save();

    //Only do the Translation to the position on the Transform layer.
    if (m_owner->hasTransformLayer()  && !isTransformLayer()) {
        ctx->translate(anchor_x*width, anchor_y*height);
    } else {
        //translate so that anchor point is the origin and apply owner's position offset
        ctx->translate(anchor_x*width + x, anchor_y*height + y);
    }

    //concatenate our transform so that is is applied to any children
    ctx->concatCTM(transform());

    // Apply perspective Transform
    if (!sublayerTransform().isIdentity()) {
        ctx->concatCTM(sublayerTransform());
    }

    //translate back from anchor
    ctx->translate(-anchor_x*width, -anchor_y*height);

    m_drawTransform = ctx->getCTMAsTransformationMatrix();

    m_accumulatedOpacity = opacity * m_opacity;

    if (drawsContent() && (width > 0) && (height > 0)) {
        //Add this layer to be z-order sorted
        sorter->addLayer(this);
    }

    //recursively paint any children
    for (int i = 0; i < children().size(); ++i) {
        children()[i]->process3DTransforms(ctx, sorter, m_accumulatedOpacity);
    }

    ctx->restore();
}

void LayerWiiU::paint3D(GraphicsContext *ctx)
{
    float width  = (float)contentBounds().width();
    float height = (float)contentBounds().height();

    ctx->save();
   // OS::Report("Paint 3D Layer(%d) %s \n",id(), name().utf8().data());
    if (masksToBounds()) {
        ctx->clip(FloatRect(m_owner->position().x(), m_owner->position().y(), width, height));
    }

    //concatenate our transform so that is is applied to any children
    ctx->setCTM(m_drawTransform);
    
    if (drawsContent())  {

        // If backface visablity is hidden and the m33 is negative the back then 
		// We set the alpha to 0 to make it invisable. 
        if (!m_doubleSided && (m_drawTransform.inverse().m33() < 0))
            ctx->setAlpha(0); 
        else 
            ctx->setAlpha(m_accumulatedOpacity);
          
        paintContents(ctx);    
     }

    ctx->restore();
}

}
#endif // USE(ACCELERATED_COMPOSITING)
